// -----------------------------------------------------------
// Code written by Michele Rossi
// Department of Information Engineering, University of Padova
//
// Contact:	rossi@dei.unipd.it
// Date: 	May 2008
// -----------------------------------------------------------

#include <time.h>                      		// define time()
#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>

// math. functions definitions
#define twopi (8*atan(1.0))
#define pi (4*atan(1.0))
#define sgn(x) (x)==0.0?0.0:x/fabs(x)
#define MAX(x,y) (x>=y ? x : y)
#define MIN(x,y) (x>=y ? y : x)
#define MIN3(x,y,z) MIN(MIN(x,y),z)

#define TRUE		1
#define FALSE		0
#define VERBOSE		TRUE		// min. verbosity level
#define DEBUG		FALSE

#define NVECT		100
#define MAXCLASSES	20

typedef unsigned int uint;

// ----------------------------------------------------------------
// Global variables
// ----------------------------------------------------------------

struct status {
	char	*allocv;
	uint	*v;
	uint	*vnew;
	uint	*r;
	long double	*beta;
	uint	k;
	long double	betak;
};

typedef struct status STATUS;

// ---------------------------------------------------------------------------------------------------------
// Begin math routines
// ---------------------------------------------------------------------------------------------------------

// *************************************************
// GAMMA: Returns the value ln[Gamma(xx)] for xx > 0
// *************************************************
long double gammln(long double xx)
{
	long double x,y,tmp,ser;
	static long double cof[6]	={76.18009172947146,-86.50532032941677,
							24.01409824083091,-1.231739572450155,
							0.1208650973866179e-2,-0.5395239384953e-5};
	int j;

	y = x = xx;
	tmp = x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser = 1.000000000190015;
	for (j=0;j<=5;j++) ser += cof[j]/++y;

	return -tmp+log(2.5066282746310005*ser/x);
}

// **************
// Returns ln(n!)
// **************
long double factln(int n)
{
	static long double a[101]; 						// is automatically initialized to zero.

	if (n < 0) { printf("Negative factorial in routine factln\n"); exit(-5); }
	if (n <= 1)	return 0.0;
	if (n <= 100)	return a[n] ? a[n] : (a[n]=gammln(n+1.0));	// If in range of table
	else			return gammln(n+1.0);						// else... out of range of table
}

// **************************************
// Calculates the factorial of an integer
// **************************************
long double fact(int n)
{
	static int ntop=4;
	static float a[33]={1.0,1.0,2.0,6.0,24.0}; 	// Fill in table up to and including 4!
							// remaining entries are calculated on demand in (*1)
	int j;

	if (n < 0)  printf("Negative factorial in routine factrl\n");
	if (n > 32) return expl(gammln(n+1.0));

	while (ntop<n) {				// (*1)
		j=ntop++;
		a[ntop]=a[j]*ntop;
	}

	return a[n];
}

// *****************************************
// Calculates the binomial coefficient (n k)
// *****************************************
long double bino(int n, int k)
{
	if (k==0) return 1.;
	return floor(0.5+expl(factln(n)-factln(k)-factln(n-k)));
}

// Returns the binomial distribution: Binomial(x,e)*p^e*(1-p)^(x-e)
long double B(int x, int e, long double p) {
	return ( bino(x,e)*(long double)powl(p,e)*(long double)powl(1.-p,x-e) );
}

void Test_Binomial(int x, int e, long double p) {

	printf("x=%d e=%d p=%g -> B(x,e,p) = %g\n", x, e, p, B(x,e,p));
}

// ************************
// ccdf exponentials type I
// ************************
//
// Computes the ccdf of a sum of exponentials x_i: y = sum_{i=1}^n x_i
// (when all n rates are equal to \alpha), i.e., returns Prob(y >= x)
long double exp_ccdf_I(uint n, long double alpha, long double x)
{
	long double	acc=0., tmp;
	uint	k;

	tmp = alpha*x;
	for (k=0; k<=n-1; k++) acc += (long double)powl(tmp,k)*expl(-tmp)/fact(k);

	return acc;
}

// *************************
// ccdf exponentials type II
// *************************
//
// Computes the ccdf of a sum of exponentials x_i: y = sum_{i=1}^n x_i
// (when all n rates \alpha_i are distinct), i.e., returns Prob(y >= x)
long double exp_ccdf_II(uint n, long double *alpha, long double x)
{
	long double	acc=0.,beta=1.,tmp;
	uint	j,k;

	// alpha is intended as a vector of lenght n from 0,1,...,n-1

	for (j=0; j<n; j++) {
		beta *= alpha[j];

		tmp = alpha[j];
		for (k=0; k<n; k++) if (k!=j) tmp *= (alpha[k]-alpha[j]);
		acc += (expl(-alpha[j]*x)/tmp);
	}
	acc *= beta;  
	return acc;
}

// **************************************************************************************
// Recursively generates all partitions of integers such that p1+p2+...+pm = n, with m<=n
// **************************************************************************************
// from: John M. Boyer, "Simple Constant Amortized Time Generation of FIxed Length Numeric Partitions,"
// Elsevier Journal of Algorithms, Number 54, pp. 31-39, 2005.
// NOTE: n is the sum, v is a vector of length ko and is used as v[1], v[2], ..., v[ko] (no v[0]!)
// USAGE:
// Gpartition(v, n, ko, ko, 1): returns all pi>=1 with sum n -> use (*1) to get sum n and all pi>=1
// Gpartition(v, n+m, ko, ko, 1): returns all pi>=1 with sum n+m -> use (*2) to get sum n and all pi>=0
void Gpartition(uint *v, uint n, uint k, uint ko, uint l)
{
	if (k==1) {
		uint i;
		v[k]=n;

		// Process or write partition somewhere
		// for (i=1; i<=ko; i++) printf("%d ", v[i]); // (*1)
		for (i=1; i<=ko; i++) printf("%d ", v[i]-1); // (*2)
		printf("\n");

	} else {
		uint h,i;

		h = (uint)floor(n/k);
		for (i=l; i<=h; i++) {
			v[k]=i;
			Gpartition(v, n-v[k],k-1,ko,v[k]);
		}
	}
}

// *********************************************************************
// Recursively lists all permutations of a vector v with length elements
// *********************************************************************
// v: original vector
// vnew: new vector on which we write permutations of v
// allocv: allocation vector; keeps track of positions of v that still have to be processed
// NOTE: v, vnew and allocv are used from 1 to length (position 0 not used!)
// pos: initially should be 1
long double permute(uint *v, uint *vnew, char *allocv, uint pos, uint length)
{
	uint	i;
	long double	acc=0.;

	if (pos==(length+1)) {
			uint j;

			// printf("pos=%d\n", pos);

			for (j=1; j<=length; j++) printf("%d ", vnew[j]);
			printf("\n");
		
			return 1;

	} else for (i=1; i<=length; i++) {

// with the following line -> obtain all permutations (possibly with repetitions)
//		if ((allocv[i]==0) {

// with the following line -> obtain all permutations BUT without repetitions!!!
		if ((allocv[i]==0) && ((v[pos]!=v[i]) || (pos==i))) {

			vnew[pos]		= v[i];
			allocv[i]		= 1;
			acc += permute(v, vnew, allocv, pos+1, length); // recursion here
			allocv[i]		= 0;
		}
	}
	return acc;												// returns the total number of generated permutations
}

// Routine derived from permute(...) above
long double permute_and_sum(STATUS *stat, uint pos, uint length)
{
	uint	i;
	long double	acc=0.;

	if (pos==(length+1)) {
			uint	ij, j;
			long double	product=1.;

			// length here is "a-1", j=1,...,a, j \neq k
			// first element of vnew is vnew[1]
			// first element of beta and r are beta[0] and r[0], respectively
			for (j=1; j<=(length+1); j++) {
				if (j!=stat->k) {
					if (j<stat->k)	ij = stat->vnew[j] - 1;			// (*3)
					else			ij = stat->vnew[j-1] - 1;		// (*4)
					product *= bino(ij+stat->r[j-1]-1,ij)*(long double)powl(stat->beta[j-1]-stat->betak,-1.*(stat->r[j-1]+ij));
				}
			}
			return product;

	} else for (i=1; i<=length; i++) {

// with the following line -> obtain all permutations
//		if ((allocv[i]==0) {

// with the following line -> obtain all permutations BUT without repetitions!!!
		if ((stat->allocv[i]==0) && ((stat->v[pos]!=stat->v[i]) || (pos==i))) {

			stat->vnew[pos]		= stat->v[i];
			stat->allocv[i]		= 1;
			acc += permute_and_sum(stat, pos+1, length); // recursion here
			stat->allocv[i]		= 0;
		}
	}
	return acc;
}

// ************************
// Function phi_{k,\ell}(x)
// ************************
// Routine derived from Gpartition(...) above
long double phi(STATUS *stat, uint n, uint k, uint ko, uint l)
{	
	
	
	long double acc=0.;

	if (k==1) {
		uint i;

		stat->v[k] = n;

// HERE: Process or write current partition somewhere
//		for (i=1; i<=ko; i++) printf("[%d] ", v[i]);
//		printf("\n");

		// reset allocation vector
		for (i=1; i<=ko; i++) stat->allocv[i] = 0;

		// process this partition (and all of its permutations, without considering repetitions)
		return permute_and_sum(stat, 1, ko);

	} else {
		uint h,i;

		h = (uint)floor(n/k);
		for (i=l; i<=h; i++) {
			stat->v[k]	= i;
			acc			+= phi(stat, n-stat->v[k], k-1, ko, stat->v[k]);
		}
		return acc;
	}
}

// **************************
// ccdf exponentials type III
// **************************
//
// Computes the ccdf of a sum of exponentials x_i: y = sum_{i=1}^n x_i
// when all some of the rates \alpha_i are equal and some are distinct,
// i.e., returns Prob(y >= x)
// a: is the number of equality classes
// beta: is a vector containing the rate for each class
// r: is a vector containing the cardinality of each class
// For math. derivations see:
// Suprasad V. Amari and Ravindra B. Misra, "Closed-Form Expressions for Distributions of Sum of Exponential Random Variables,"
// IEEE Transactions on Reliability, Vol. 46, No. 4, pp. 519-522, December 1997.
long double exp_ccdf_III(uint a, long double *beta, uint *r, long double x)
{
	static	uint	*v=NULL;
	static	uint	*vnew=NULL;
	static	char	*allocv=NULL;
	static	STATUS	*stat=NULL;

	uint	k, ell, j;
	long double	c=1., acc=0.;

	if (a>(MAXCLASSES-1)) { printf("ERROR: increase MAXCLASSES in function exp_ccdf_III(...)"); exit(-4); }

	// allocate vectors that will be used for this computation
	// use static vectors for increased speed when used intensively
	if (v==NULL) {
		v		= (uint *)malloc(sizeof(uint)*NVECT);
		vnew	= (uint *)malloc(sizeof(uint)*MAXCLASSES);
		allocv	= (char *)malloc(sizeof(char)*MAXCLASSES);
		stat	= (STATUS *)malloc(sizeof(STATUS));

		stat->v			= v;
		stat->vnew		= vnew;
		stat->allocv	= allocv;
	}

	stat->r		= r;
	stat->beta	= beta;

	for (k=0; k<a; k++) {

		c *= (long double)powl(beta[k], r[k]);	// multiplicative constant

		// optimization treat the cases r[k]>1 and ==1 separately
		// as when r[k]==1 further optimizations are possible
		if (r[k]>1) {
			
			for (ell=1; ell<=r[k]; ell++) {		// sum over all partitions of \sum_{j=1,\neq k}^a i_j = l-1+a-1 with i_j>=1
												// i_j are reduced to i_j>=0 and \sum_{j=1,\neq k}^a i_j = l-1 in (*3) and (*4) above (subtracting 1)
					stat->k		= k + 1;		// we add one as it will be used on vectors from 1,2,... instead of 0,1,...
					stat->betak = beta[k];

					// following if statement to replace instruction acc +=pow(-1., ell-1)* ... ;
					if ((ell-1)%2==0)	acc += (phi(stat,ell-1+a-1,a-1,a-1,1)*exp_ccdf_I(r[k]-ell+1,beta[k],x)/(long double)powl(beta[k],r[k]-ell+1));
					else				acc -= (phi(stat,ell-1+a-1,a-1,a-1,1)*exp_ccdf_I(r[k]-ell+1,beta[k],x)/powl(beta[k],r[k]-ell+1));
			}

		} else if (r[k]==1) {
			long double product=1.;

			for (j=0; j<a; j++) if (j!=k) product *= (long double)powl(beta[j]-beta[k], -1.*r[j]);
			acc += product*exp_ccdf_I(r[k],beta[k],x)/(long double)powl(beta[k],r[k]);
		}
	}
	acc *= c;
	
	return acc;
}

// ***************************************
// Split rate vector into equality classes
// ***************************************
// split the input rates vector \alpha into equality classes
// store into vectors beta and r the rate for each class
// and its cardinality, respectively. The function returns
// a, i.e., the number of equality classes found
uint split_into_classes(uint n, long double *alpha, long double *beta, uint *r)
{
	uint	a, j, k;
	char	found;

	a		= 1;																// number of equality classes
	beta[0] = alpha[0];
	r[0]	= 1;

	for (k=1; k<n; k++) {
		found = 0;
		for (j=0; j<a; j++) if (alpha[k]==beta[j]) { found=1; r[j]++; }			// increase occupancy for class j

		if (!found) {
				r[a]	= 1;
				beta[a]	= alpha[k];												// create a new equality class with value alpha[k]
				a++;															// increase number of equality classes
		}
	}

	return(a);
}

// ***************************
// ccdf of sum of exponentials
// ***************************
//
// Entry point for the calculation of the ccdf of sums of exponentials
// y = \sum_{i=1}^n x_i, with generic rates \alpha_i
long double exp_ccdf(uint n, long double *alpha, long double x)
{
	static long double	*beta=NULL;
	static uint		*r=NULL;
	long double			prob;
	uint			a;

	if (n>NVECT) { printf("ERROR: increase NVECT in function exp_ccdf(...)\n"); exit(-2); }

	// optimization: in order not to allocate and de-allocate vectors at each use
	// we use static variables here, this should speed up things when the function
	// is called multiple times
	if (beta==NULL) {
		beta	= (long double *)malloc(sizeof(long double)*NVECT);
		r		= (uint *)malloc(sizeof(uint)*NVECT);
	}

	// returns the number of equality classes
	// and fills vectors beta and r, see explanations above
	a = split_into_classes(n, alpha, beta, r);

	if (DEBUG) {
		uint i;

		printf("Equality classes=%d\n", a);
		for (i=0; i<a; i++) printf("beta[%d]=%Lf, cardinality=%d\n", i, beta[i], r[i]);
		printf("\n");
	}

	if (a==1)		prob = exp_ccdf_I(n, alpha[0], x);
	else if (a==n)	prob = exp_ccdf_II(n, alpha, x);
	else if (a<n)	prob = exp_ccdf_III(a, beta, r, x);
	else { printf("ERROR: check equality classes in function exp_ccdf(...)\n"); exit(-3); }

	return prob;
}

// ---------------------------------------------------------------------------------------------------------
// End math routines
// ---------------------------------------------------------------------------------------------------------
